local RW = LibStub("AceAddon-3.0"):GetAddon("Raid Watch")
if not RW then return end
local L = LibStub("AceLocale-3.0"):GetLocale("RW2-Base")
local MapData = LibStub("LibMapData-1.0")
local LSM = LibStub("LibSharedMedia-3.0")


local Utils = {}
RW.Utils = Utils

local Party
local math_modf = math.modf

function Utils:InTable(t, entry)
	for i, v in pairs(t) do
		if v == entry then return true end
	end
end

function Utils:InTable2(t, key, entry)
	for i, v in pairs(t) do
		if v[key] == entry then return true end
	end
end

function Utils:GetTableData(t, key, entry)
	for i, v in pairs(t) do
		if v[key] == entry then return v end
	end
end

function Utils:GetIndexForvalue(t, value)
	for i, v in pairs(t) do
		if v == value then return i end
	end
end

function Utils:NrElements(t)
	local count = 0
	for i, v in pairs(t) do
		count = count + 1
	end
	return count
end

function Utils:TRemove(t, entry)
	for i, v in pairs(t) do
		if v == entry then 
			tremove(t, i)
		end
	end
end

-- Unbound CopyTable
function Utils:CopyTableUnlimited(t, d)
	local copy = {}
	if type(t) ~= "table" then return t end
	for i,v in pairs(t) do
		if type(v) ~= "table" then
			copy[i] = v
		else
			if d == nil then
				d = {}
			end
			d[t] = copy
			if d[v] then
				copy[i] = d[v] -- we already copied this table. reuse the copy.
			else
				copy[i] = self:CopyTableUnlimited(v,d)
			end
		end
	end
	return copy
end

function Utils:CopyTable(t, d)
	if (d  or 0) > 5 then return nil end
	local new = {}
	for i, v in pairs(t) do
		if type(v) == "table" then
			new[i] = self:CopyTable(v, (d or 0) + 1)
		else
			new[i] = v
		end
	end
	return new
end

function Utils:CopyTableShallow(s, d)
	if type(s) ~= "table" or type(d) ~= "table" then return end
	
	for i, v in pairs(s) do
		if type(v) == "table" then
			self:CopyTableShallow(v, d[i])
		else
			d[i] = v
		end
	end
end

function Utils:TimeToStr(time)
	if type(time) ~= "number" then return time end
	time = math.floor(time)
	local t
	if time < 60 then
		return L.COMBATTIME_SEC:format(time)
	elseif time % 60 == 0 then
		return L.COMBATTIME_MIN:format(time/60)
	else
		return L.COMBATTIME:format(time/60, time % 60)
	end
end

function Utils:TimeToStringShort(t)
	if type(t) ~= "number" then return t end
	if t <= 10 then
		return ("%.1fs"):format(t)
	elseif t <= 60 then
		return ("%ds"):format(t)
	else
		return ("%0.2d:%0.2d"):format(t/60, math.fmod(t, 60))
	end
end

function Utils:RGBConvert(r, g, b, a)
	return r/255, g/255, b/255, (a and a/255 or 1)
end

function Utils:Lerp(s, e, a)
	if not s or not e or not a then return 0 end
	return s + ((e - s) * a)
end

function Utils:ColorNames(str)
	if not Party then Party = RW:GetPlugin("Party") end
	str = str:gsub(">([%S\s-]+)<", function(name)
		if Party.group[name] then
			local color = RAID_CLASS_COLORS[Party.group[name].class]
			if color then
				name = ("|cff%.2x%.2x%.2x%s|r"):format(color.r * 255, color.g * 255, color.b * 255, name)
			end
		end
		return name
	end)
	return str
end

-- A bit of hack to solve if the first argument is correct but the rest is not, 
-- only for timer and message formating realy.
function Utils:Format(text, ...)
	local ok, str = pcall(format, text, ...)
	if not ok then
		local first = select(1, ...)
		ok, str = pcall(format, text, first, L.UNKNOWN, 0)
	end
	if not ok then
		local first = select(1, ...)
		ok, str = pcall(format, text, first, 0)
	end
	return ok and str or text:gsub("(%%.)", L.UNKNOWN)
end

Utils.SpellName = setmetatable({}, {
	__index = function(t, key)
		if type(key) ~= "number" then return key end
		local name = GetSpellInfo(key)
		if not name then 
			print(("Cannot fetch spell name for %q"):format(key)) 
			return key
		end
		return name		
	end}
)

do
	local playerSpells
	function Utils:GetRandomSpell()
		if not playerSpells then
			playerSpells = {}
			local i = 1
			while true do
				local spellName = GetSpellName(i, BOOKTYPE_SPELL)
				if not spellName then break end
				playerSpells[#playerSpells + 1] = spellName
				i = i + 1
			end
		end
		local name, rank, icon = GetSpellInfo(playerSpells[math.random(1, #playerSpells)])
		return name, rank, icon
	end
end

-- Tekkub's ColorGradient() http://www.wowwiki.com/ColorGradient
function Utils:ColorGradient(perc, r1, g1, b1, r2, g2, b2, r3, g3, b3)
	if perc >= 1 then return r3, g3, b3 elseif perc <= 0 then return r1, g1, b1 end
	local segment, relperc = math_modf(perc*2)
	if segment == 1 then r1, g1, b1, r2, g2, b2 = r2, g2, b2, r3, g3, b3 end
	return r1 + (r2-r1)*relperc, g1 + (g2-g1)*relperc, b1 + (b2-b1)*relperc
end

function Utils:ColorGradientSimple(perc, r1, g1, b1, r2, g2, b2)
	if perc >= 1 then return r2, g2, b2
	elseif perc <= 0 then return r1, g1, b1
	end
	return r1 + (r2-r1)*perc, g1 + (g2 - g1)*perc, b1 + (b2 - b1)*perc
end

-----------------------------
-- Party functions
function Utils:IsHeroic()
	local name, instanceType, difficultyIndex, difficultyName, maxPlayers, playerDifficulty, isDynamic = GetInstanceInfo()
	if (instanceType == "party" and difficultyIndex == 2) or difficultyIndex > 2 then return true end
	if instanceType == "none" then return false end
end

function Utils:InstanceSize()
	local _, instanceType, difficulty, _, maxPlayers = GetInstanceInfo()
	if instanceType == "none" then return 10 end
	return maxPlayers
end

-- Constants
Utils.Constants = {}
Utils.Constants.FontFlagList = {
	["NONE"] = L.NONE,
	["OUTLINE"] = L.OUTLINE,
	["THICKOUTLINE"] = L.THICK_OUTLINE,
	["MONOCHROME"] = L.MONOCHROME,
}

Utils.Constants.UpDownList = {
	["TOP"] = L.TOP,
	["BOTTOM"] = L.BOTTOM,
}

Utils.Constants.RaidIconNameToNumber = {
	STAR	= 1,
	CIRCLE 	= 2,
	DIAMOND	= 3,
	TRIANGLE = 4,
	MOON 	= 5,
	SQUARE 	= 6,
	CROSS 	= 7,
	SKULL 	= 8,
}

Utils.Constants.IconCoords = {
	STAR		= {0, 		0.25,	0, 		0.25},
	CIRCLE 		= {0.25,	0.5,	0,		0.25},
	DIAMOND		= {0.5, 	0.75,	0,		0.25},
	TRIANGLE 	= {0.75,	1,		0,		0.25},
	MOON 		= {0,		0.25,	0.25,	0.5},
	SQUARE 		= {0.25,	0.5,	0.25,	0.5},
	CROSS 		= {0.5,		0.75,	0.25,	0.5},
	SKULL 		= {0.75,	1,		0.25,	0.5},
}



Utils.Constants.ButtonTextures = {
	["CloseUp"] = [[Interface\AddOns\RaidWatch\Media\Icons\close_up.tga]],
	["CloseDown"] = [[Interface\AddOns\RaidWatch\Media\Icons\close_down.tga]],
	
	["DKPUp"] = [[Interface\AddOns\RaidWatch\Media\Icons\dkp_up.tga]],
	["DKPDown"] = [[Interface\AddOns\RaidWatch\Media\Icons\dkp_down.tga]],
	
	["RangeCheckUp"] = [[Interface\AddOns\RaidWatch\Media\Icons\rangecheck_up.tga]],
	["RangeCheckDown"] = [[Interface\AddOns\RaidWatch\Media\Icons\rangecheck_down.tga]],
	
	["ConfigUp"] = [[Interface\AddOns\RaidWatch\Media\Icons\config_up.tga]],
	["ConfigDown"] = [[Interface\AddOns\RaidWatch\Media\Icons\config_down.tga]],
	
	["FilterUp"] = [[Interface\AddOns\RaidWatch\Media\Icons\filter_up.tga]],
	["FilterDown"] = [[Interface\AddOns\RaidWatch\Media\Icons\filter_down.tga]],
	
	["RefreshUp"] = [[Interface\AddOns\RaidWatch\Media\Icons\refresh_up.tga]],
	["RefreshDown"] = [[Interface\AddOns\RaidWatch\Media\Icons\refresh_down.tga]],
	
	["UpPointerUp"] = [[Interface\AddOns\RaidWatch\Media\Icons\uppointer_up.tga]],
	["UpPointerDown"] = [[Interface\AddOns\RaidWatch\Media\Icons\uppointer_down.tga]],
	
	["DownPointerUp"] = [[Interface\AddOns\RaidWatch\Media\Icons\downpointer_up.tga]],
	["DownPointerDown"] = [[Interface\AddOns\RaidWatch\Media\Icons\downpointer_down.tga]],
	
	["LockUp"] = [[Interface\AddOns\RaidWatch\Media\Icons\lock_up.tga]],
	["LockDown"] = [[Interface\AddOns\RaidWatch\Media\Icons\lock_down.tga]],
	
	["Highlight"] = [[Interface\AddOns\RaidWatch\Media\Icons\button_highlight.tga]],
}

-----------------------------
-- Frames
function Utils:CreateAnchor(name, parent, text)
	local frame = CreateFrame("Frame", name or nil, parent or nil)
	frame:SetPoint("CENTER")
	frame:SetWidth(200)
	frame:SetHeight(25)
	frame:EnableMouse(true)
	frame:RegisterForDrag("LeftButton")
	frame:SetFrameStrata("BACKGROUND")
	frame:SetMovable(true)

	local bg = frame:CreateTexture(nil, "PARENT")
	bg:SetAllPoints(frame)
	bg:SetTexture(1, 1, 1, 1)
	bg:SetGradient("HORIZONTAL", 0.39, 0.55, 0.87, 0.1, 0.1, 0.1)
	frame.background = bg
	
	local str = frame:CreateFontString(name and name.."Text", "ARTWORK")
	str:SetFont(LibStub("LibSharedMedia-3.0"):Fetch("font", "accid"), 16, "OUTLINE")
	str:SetAllPoints()
	str:SetText(text)

	local button = self:CreateButton("Close")
	button:SetParent(frame)	
	button:SetPoint("RIGHT", -5, 0)
	button:AddClickHandler(function(self)
		self:GetParent():Hide()
	end)
		
	frame.hidebutton = button
	frame.str = str
	frame:Hide()
	return frame
end

local barCounter = 0
function Utils:CreateBar(rightClickFunc)
	local frame = CreateFrame("Frame", "RW_Bar"..barCounter, UIParent)
	frame.ID = barCounter
	barCounter = barCounter + 1
	local bar = CreateFrame("StatusBar", frame:GetName().."Bar", frame)
	local timeText = bar:CreateFontString(frame:GetName().."TimeText", "ARTWORK")
	local text = bar:CreateFontString(frame:GetName().."Text", "ARTOWRK")
	local icon = frame:CreateTexture(frame:GetName().."Icon", "ARTWORK")
	
	frame:SetPoint("CENTER")
	frame:SetWidth(255)
	bar:SetPoint("BOTTOMRIGHT", frame, "BOTTOMRIGHT", -3, 3)
	bar:SetPoint("TOPRIGHT", frame, "TOPRIGHT", -3, -3)
	icon:SetPoint("TOPLEFT", frame, "TOPLEFT", 3, -3)
	icon:SetPoint("BOTTOMRIGHT", bar, "BOTTOMLEFT", -2, 0)
	timeText:SetPoint("RIGHT", bar, "RIGHT", -4, 0)
	text:SetPoint("LEFT", bar, "LEFT", 2, 0)
	text:SetPoint("RIGHT", timeText, "LEFT")
	
	bar:SetMinMaxValues(0,1)
	
	timeText:SetJustifyH("LEFT")
	timeText:SetWidth(45)
	
	frame.timeText = timeText
	frame.text = text
	frame.bar = bar
	frame.icon = icon
	
	frame.rightClickfuc = rightClickFunc
	
	frame:SetScript("OnEvent", function(self, event, arg1, arg2)
		if arg1 == "LSHIFT" or arg1 == "RSHIFT" then
			if arg2 == 1 then
				frame:EnableMouse(true)
				self:SetScript("OnMouseDown", self.rightClickfuc)
			else
				frame:EnableMouse(false)
				self:SetScript("OnMouseDown", nil)
			end
		end
	end)
	frame:Hide()
	return frame
end

do
	local function ShowTooltip(self)
		GameTooltip:SetOwner(self, "NONE")
		if self.TT_Title then GameTooltip:AddLine(self.TT_Title, nil, nil, nil, true) end
		if self.TT_Text then GameTooltip:AddLine(self.TT_Text, nil, nil, nil, true) end
		GameTooltip:Show()
	end
	
	local function HideTooltip(self)
		GameTooltip:Hide()
	end
	
	local function buttonClick(self, button, IsDown)		
		if not IsDown then			
			for i, handler in pairs(self.clickhandlers) do
				handler(self, button, IsDown)
			end
		end
	end
	
	function Utils:AddTooltip(frame, text, title)		
		frame.TT_Title = title
		frame.TT_Text = text
		frame:HookScript("OnEnter", ShowTooltip)
		frame:HookScript("OnLeave", HideTooltip)
	end
	
	local prot = {}
	local buttonProt = {}
	local backdrop = {}
	local AllWindows = {}
	
	function prot:SetBorder(border, width, color)
		backdrop.edgeFile = border
		backdrop.edgeSize = width
		self:SetBackdrop(backdrop)
		self:SetBackdropBorderColor(unpack(color))
	end
	
	function prot:SetInsets(inset)
		self.title:SetPoint("TOPLEFT", inset, -inset)
		self.title:SetPoint("TOPRIGHT", -inset, -inset)
		self.content:SetPoint("BOTTOMRIGHT", -inset, inset)
	end
	
	function prot:SetContentHeight(height)
		self:SetHeight(height + self.title:GetHeight())
		self.contentheight = height
	end
	
	function prot:SetTitleHeight(height)
		self.title:SetHeight(height)
		self:SetContentHeight(self.contentheight)
	end
	
	function prot:SetTitle(str)
		self.title.string:SetText(str)
	end
	
	function prot:SetTitleFont(font, size, flags, color)
		self.title.string:SetFont(font, size, flags)
		self.title.string:SetTextColor(unpack(color))
	end
	
	function prot:SetTitleColor(useGradient, ...)
		r, g, b, a = ...
		if useGradient then
			self.title.background:SetGradientAlpha("HORIZONTAL", r, g, b, a, 0, 0, 0, a * 2.0)
		else
			self.title.background:SetVertexColor(r, g, b, a)
		end
	end
	
	function prot:SetBackgroundColor(...)
		self.content.background:SetTexture(...)
	end
	
	function prot:AddTitleButton(texture, iconsize, fade)
		self.buttons = self.buttons or {}
		local button = Utils:CreateButton(texture, iconsize)
		button:SetPoint("RIGHT", self.lastbutton or self.title, self.lastbutton and "LEFT" or "RIGHT", -4, 0)
		button:SetParent(self.title)
		button.fade = fade
		if fade then button:SetAlpha(0) button:Hide() end
		
		self.title.string:SetPoint("RIGHT", button, "LEFT")		
		
		tinsert(self.buttons, button)
		self.lastbutton = button
		
		return button
	end
	
	function prot:AddPulloutButton(texture, iconsize)
		self.pulloutbuttons = self.pulloutbuttons or {}
		local button = Utils:CreateButton(texture, iconsize)
		button:SetPoint("RIGHT", self.lastpulloutbutton or self.title.pullout, self.lastpulloutbutton and "LEFT" or "RIGHT", -4, 0)
		button:SetParent(self.title.pullout)
		
		tinsert(self.pulloutbuttons, button)
		self.lastpulloutbutton = button
		
		return button
	end
	
	function prot:HideContent()
		self:SetHeight(self.title:GetHeight())
		self.contentheight = 0
	end
	
	function prot:SetContentSide(side)
		if side == "TOP" then
			self.content:ClearAllPoints()
			self.title:ClearAllPoints()			
			self.title:SetPoint("BOTTOMLEFT")
			self.title:SetPoint("BOTTOMRIGHT")
			self.content:SetPoint("BOTTOMLEFT", self.title, "TOPLEFT")
			self.content:SetPoint("TOPRIGHT", self, "TOPRIGHT")						
		else
			self.content:ClearAllPoints()
			self.title:ClearAllPoints()			
			self.title:SetPoint("TOPLEFT")
			self.title:SetPoint("TOPRIGHT")
			self.content:SetPoint("TOPLEFT", self.title, "BOTTOMLEFT")
			self.content:SetPoint("BOTTOMRIGHT", self, "BOTTOMRIGHT")	
		end
	end
	
	function prot:SetPulloutSide(side)
		if side == "TOP" then
			self.title.pullout:ClearAllPoints()			
			self.title.pullout:SetPoint("BOTTOMLEFT", self.title, "TOPLEFT")
			self.title.pullout:SetPoint("BOTTOMRIGHT", self.title, "TOPRIGHT")
			self.title.pullout.line:ClearAllPoints()
			self.title.pullout.line:SetPoint("BOTTOMLEFT")
		    self.title.pullout.line:SetPoint("BOTTOMRIGHT")
		else
			self.title.pullout:ClearAllPoints()			
			self.title.pullout:SetPoint("TOPLEFT", self.title, "BOTTOMLEFT")
			self.title.pullout:SetPoint("TOPRIGHT", self.title, "BOTTOMRIGHT")
			self.title.pullout.line:ClearAllPoints()
			self.title.pullout.line:SetPoint("TOPLEFT")
			self.title.pullout.line:SetPoint("TOPRIGHT")
		end
	end
	
	function prot:AddLeftIcon()
		local icon = CreateFrame("Frame", nil, self.title)
		icon:SetPoint("TOPLEFT", 3, -3)		
		icon:SetPoint("BOTTOMLEFT", 3, 3)
		self.title.lefticon = icon
		
		local t = icon:CreateTexture(nil, "OVERLAY")	
		t:SetPoint("TOPLEFT", 2, -2)
		t:SetPoint("BOTTOMRIGHT", -2, 2)
		t:SetTexCoord(0.07, 0.93, 0.07, 0.93)				
		icon.texture = t
		
		t = icon:CreateTexture(nil, "ARTWORK")
		t:SetPoint("TOPLEFT")
		t:SetPoint("BOTTOMRIGHT")
		t:SetTexture(0, 0, 0, 1)
		icon.background = t
	end
	
	function prot:ShowLeftIcon(texture)
		if not self.title.lefticon then self:AddLeftIcon() end
		self.title.lefticon:Show()
		self.title.lefticon.texture:SetTexture(texture)
		self.title.lefticon:SetWidth(self.title:GetHeight() - 6)
		self.title.string:SetPoint("LEFT", self.title.lefticon, "RIGHT", 3, 0)
	end
	
	function prot:HideLeftIcon()
		if not self.title.lefticon then return end
		self.title.lefticon:Hide()
		self.title.string:SetPoint("LEFT", 2, 0)
	end
	
	function Utils:CreateFrame(id, parent)
		local frame = CreateFrame("Frame", id, parent)
		local background = frame:CreateTexture(nil, "OVERLAY")
		background:SetAllPoints()
		background:SetTexture(1, 1, 1, 1)
		frame.background = background
		
		return frame
	end
	
	local function ShowPullout(self)
		if self.pullout then
			Utils.FadeIn(self.pullout)
		end
		for i, button in pairs(self.window.buttons) do
			if button.fade then
				Utils.FadeIn(button)
			end
		end
	end
	
	local function HidePullout(self)
		for i, button in pairs(self.window.buttons) do
			if button.fade then
				Utils.FadeOut(button)
			end
		end
		if self.pullout then
			if GetMouseFocus() == self.pullout then
				return
			end
			Utils.FadeOut(self.pullout)
		end
	end
		
	function Utils:CreateWindow(windowID, movable, moveStart, moveEnd, noPullout)
		local r, g, b, a
		local db = RW.db.profile
		local frame = CreateFrame("Frame", "RW_Window_"..windowID, UIParent)
		
		local title = self:CreateFrame("RW_Window_Title"..windowID, frame)
		title:SetFrameLevel(title:GetFrameLevel() + 10)
		title:SetPoint("TOPLEFT")
		title:SetPoint("TOPRIGHT")
		title:SetHeight(db.WindowTitleHeight)	
		title:SetScript("OnEnter", ShowPullout)
		title:SetScript("OnLeave", HidePullout)		
		r, g, b, a = unpack(db.WindowTitleColor)
		title.background:SetGradientAlpha("HORIZONTAL", r, g, b, a, 0, 0, 0, a * 2)
		title.window = frame
		
		local titlestring = title:CreateFontString(nil, "OVERLAY", "GameFontNormal")
		titlestring:SetPoint("LEFT", 2, 0)
		titlestring:SetShadowOffset(1, -1)
		titlestring:SetJustifyH("LEFT")
		titlestring:SetFont(LSM:Fetch("font", db.WindowFont), db.WindowFontSize, db.WindowFontFlags)
		titlestring:SetTextColor(unpack(db.WindowFontColor))
		title.string = titlestring
		
		if not noPullout then
			local pullout = self:CreateFrame("RW_Window_Pullout"..windowID, title)
			pullout:SetPoint("BOTTOMLEFT", title, "TOPLEFT")
			pullout:SetPoint("BOTTOMRIGHT", title, "TOPRIGHT")
			pullout:SetHeight(db.WindowTitleHeight)
			pullout:SetScript("OnEnter", Utils.FadeIn)
			pullout:SetScript("OnLeave", Utils.FadeOut)
			pullout.background:SetGradientAlpha("HORIZONTAL", r, g, b, a, 0, 0, 0, a * 2)
			pullout:SetClampedToScreen(true)
			pullout:EnableMouse(true)
			pullout:SetAlpha(0)
			pullout:Hide()
			title.pullout = pullout
			
			local line = self:CreateFrame("RW_Window_Line"..windowID, pullout)
			line:SetFrameLevel(pullout:GetFrameLevel() + 1)
			line:SetPoint("BOTTOMLEFT")
			line:SetPoint("BOTTOMRIGHT")
			line:SetHeight(2)		
			line.background:SetTexture(0, 0, 0, 1)
			pullout.line = line
		end
		
		local content = self:CreateFrame("RW_Window_Content"..windowID, frame)
		content:SetFrameLevel(content:GetFrameLevel() + 10)
		content:SetPoint("TOPLEFT", title, "BOTTOMLEFT")
		content:SetPoint("BOTTOMRIGHT", frame)
		content.background:SetTexture(0, 0, 0, 0.5)
		
		frame.title = title
		frame.content = content
		
		for i, func in pairs(prot) do
			frame[i] = func
		end
		
		title:EnableMouse(true)
		if movable then
			frame:SetMovable(true)				
			title:RegisterForDrag("LeftButton")
			title:SetScript("OnDragStart", moveStart or (function() frame:StartMoving() end))
			title:SetScript("OnDragStop", moveEnd or (function() frame:StopMovingOrSizing() end))
			if pullout then
				pullout:RegisterForDrag("LeftButton")
				pullout:SetScript("OnDragStart", moveStart or (function() frame:StartMoving() end))
				pullout:SetScript("OnDragStop", moveEnd or (function() frame:StopMovingOrSizing() end))
			end
		end
		
		tinsert(AllWindows, frame)
		
		return frame
	end
	
	local function ButtonEnter(self)
		local parent = self:GetParent()
		if not parent then return end
		local script = parent:GetScript("OnEnter")
		if not script then return end
		script(parent)
	end
	
	local function ButtonLeave(self)
		local parent = self:GetParent()
		if not parent then return end
		local script = parent:GetScript("OnLeave")
		if not script then return end
		script(parent)
	end
	
	function Utils:CreateButton(texture, iconsize)
		self.buttons = self.buttons or {}
		local button = CreateFrame("Button", "Utils_Button_"..#self.buttons, UIParent)
						
		button:SetScript("OnClick", buttonClick)
		button:HookScript("OnEnter", ButtonEnter)
		button:HookScript("OnLeave", ButtonLeave)
		button:RegisterForClicks("LeftButtonUp", "LeftButtonDown")
		
		local t = button:CreateTexture(nil, "ARTWORK")
		t:SetAllPoints()
		t:SetTexture(Utils.Constants.ButtonTextures[texture.."Up"])
		button.UpTexture = t
		button:SetNormalTexture(t)
		
		t = button:CreateTexture(nil, "ARTWORK")
		t:SetAllPoints()
		t:SetTexture(Utils.Constants.ButtonTextures[texture.."Down"])
		button.DownTexture = t
		button:SetPushedTexture(t)
		
		t = button:CreateTexture(nil, "ARTWORK")
		t:SetAllPoints()
		t:SetTexture(Utils.Constants.ButtonTextures["Highlight"])
		button.HighlightTexture = t
		button:SetHighlightTexture(t, "ADD")
		
		for i, func in pairs(buttonProt) do
			button[i] = func
		end
		
		button:ApplyStyle()
		button:SetHeight(iconsize or RW.db.profile.WindowButtonSize)
		button:SetWidth(iconsize or RW.db.profile.WindowButtonSize)
		
		tinsert(self.buttons, button)
		
		return button
	end
	
	function buttonProt:AddClickHandler(handler)
		self.clickhandlers = self.clickhandlers or {}
		tinsert(self.clickhandlers, handler)
	end
	
	function buttonProt:ApplyStyle()
		local db = RW.db.profile
		self:SetWidth(db.WindowButtonSize)
		self:SetHeight(db.WindowButtonSize)
		self.UpTexture:SetVertexColor(unpack(db.WindowButtonColor))
		self.DownTexture:SetVertexColor(unpack(db.WindowButtonColor))
	end
	
	function Utils:ApplyWindowSettings()
		for i, window in pairs(AllWindows) do
			local db = RW.db.profile
			window:SetTitleHeight(db.WindowTitleHeight)
			window:SetTitleFont(LSM:Fetch("font", db.WindowFont), db.WindowFontSize, db.WindowFontFlags, db.WindowFontColor)
			window:SetTitleColor(db.WindowUseGradient, unpack(db.WindowTitleColor))
			if window.title.lefticon then
				window.title.lefticon:SetWidth(db.WindowTitleHeight - 4)
			end
			if window.title.pullout then
				local r, g, b, a = unpack(db.WindowTitleColor)
				window.title.pullout.background:SetGradientAlpha("HORIZONTAL", r, g, b, a, 0, 0, 0, a * 2)
			end
			
			for j, button in pairs(window.buttons) do
				button:ApplyStyle()
			end
			if window.pulloutbuttons then
				for j, button in pairs(window.pulloutbuttons) do
					button:ApplyStyle()
				end
			end
		end
	end
end


----------
-- Map distance stuff
do
	local CurrentZone, CurrentFloor, CurrentW, CurrentH
	MapData:RegisterCallback("MapChanged", function(event, map, floor, w, h)
		CurrentZone = map
		CurrentFloor = floor
		CurrentW = w
		CurrentH = h
	end)
	
	function Utils:GetDistanceAndAngle(x, y)
		return MapData:DistanceAndDirection(CurrentZone, CurrentFloor, x, y)
	end
	
	function Utils:GetPointInYards(x, y)
		return MapData:PointToYards (CurrentZone, CurrentFloor, x, y)
	end
	
	function Utils:GetDistanceAndAngleToUnit(uID)
		if not UnitIsConnected(uID) then return end
		local tx, ty = GetPlayerMapPosition(uID)
		return self:GetDistanceAndAngle(tx, ty)
	end
	
	function Utils:IsUnitInRange(uID, range)
		local dist = self:GetDistanceAndAngleToUnit(uID)
		if dist then return dist <= range end
		return false
	end
end


-- OnUpdate Stuff -------------------
-------------------------------------
do
	local OnUpdateFuncs = {}
	local FuncElapsed = {}
	local UpdateFrame = CreateFrame("Frame")
	UpdateFrame:SetScript("OnUpdate", function(self, elapsed)
		for module, functions in pairs(OnUpdateFuncs) do
			for func, data in pairs(functions) do
				if data.updateRate then
					data.elapsed = data.elapsed + elapsed
					if data.elapsed > data.updateRate then
						func(module, data.elapsed)
						data.elapsed = 0
					end
				else
					func(module, elapsed)
				end
			end
		end
	end)	
	
	function Utils:RegisterOnUpdate(module, func, updateRate)
		OnUpdateFuncs[module] = OnUpdateFuncs[module] or {}
		OnUpdateFuncs[module][func] = {updateRate = updateRate, elapsed = 0}
	end
	
	function Utils:UnregisterOnUpdate(module, func)
		OnUpdateFuncs[module][func] = nil
	end
	
	local function Fader(self, elapsed)
		local currentAlpha = self:GetAlpha()
		if self.fadingIn then
			self:Show()
			if currentAlpha >= 1 then 
				Utils:UnregisterOnUpdate(self, Fader)
				return
			end
			self:SetAlpha(self:GetAlpha() + elapsed * 2)
		else
			if currentAlpha <= 0 then 
				self:Hide()
				Utils:UnregisterOnUpdate(self, Fader)
				return
			end
			self:SetAlpha(self:GetAlpha() - elapsed * 2)
		end
	end
	
	function Utils.FadeIn(frame)
		frame.fadingIn = true
		Utils:RegisterOnUpdate(frame, Fader)
	end
	
	function Utils.FadeOut(frame)
		frame.fadingIn = nil
		Utils:RegisterOnUpdate(frame, Fader)
	end
end

-- Dungeon Stuff ------------------------
-----------------------------------------
function Utils:IsHeroic()
	-- Special case when not inside an instance
	if not IsInInstance() then
		if GetNumRaidMembers() > 0 then
			local diff = GetRaidDifficulty()
			if diff == 3 or diff == 4 then return true else return false end
		else
			local diff = GetDungeonDifficulty()
			if diff == 2 then return true else return false end
		end
	end
	_, _, difficultyIndex, _, maxPlayers, dynamicHeroic, isDynamic = GetInstanceInfo()
	
	-- If its a dynamic instance then the dynamicHeroic variable can be used 1 = heroic, 0 = normal
	if isDynamic then
		if dynamicHeroic == 1 then
			return true
		else
			return false
		end
	end
	
	-- If inside party dungeon
	if maxPlayers == 5 and difficultyIndex == 2 then
		return true
	-- If dificultyindes is above 2 and in a raid instance
	elseif difficultyIndex > 2 then
		return true
	end
	
	return false
end

function Utils:InstanceSize()
	-- Special case when not inside an instance
	if not IsInInstance() then
		if GetNumRaidMembers() > 0 then
			local size = GetRaidDifficulty()
			if size == 1 or size == 3 then return 10 else return 25 end
		else
			return 5
		end
	end
	_, _, _, _, maxPlayers = GetInstanceInfo()
	
	-- maxPlayers == 0 when outside of an instance, then return 25 for raid and 5 for party mods
	if maxPlayers == 0 then
		if self.type == "raid" then
			return 25
		elseif self.type == "party" then
			return 5
		end
	end
	
	return maxPlayers
end